/**
 * Defines the functionality of the master module in the airline
 * reservation system.
 **/

#ifndef FLIGHT_PLANNER_H
#define FLIGHT_PLANNER_H

#include <systemc.h>
#include <tlm.h>

#include "bus_types.h"
#include "basic_module.h"
#include "basic_protocol.h"
#include "bounded_queue.h"
#include "bounded_set.h"

using namespace basic_protocol;
using tlm::tlm_tag;
using user::bounded_queue;

#ifdef PROPERTY_1
// Verification 
class property1;
#endif

#ifdef PROPERTY_2
#include "property2.h"
#endif


namespace airline {
  
  class flight_planner : 
    public sc_module,
    public user::basic_module {


#ifdef PROPERTY_1
    // Verification
    // Property 1 needs access to private data fields
    friend class ::property1;
#endif
    
  public:

    // The constructor
    flight_planner(sc_module_name module_name, unsigned int q_size);
    
    SC_HAS_PROCESS( flight_planner );

    nb_in_port_t  in_from_master;
    nb_out_port_t out_to_master; // Output port to MASTER
    sc_in_clk clk;
    
    virtual void before_end_of_elaboration();

  private:

    
    // Consume new requests from the channel
    void receive_new_requests();

    // Push connectivity requests in the appropriate queue
    void request_connectivity();

    // Find connecting flights
    void find_connectivity();

    // Send back completed requests to the master
    void send_finished_requests();

    // Handles special requests (delete, cancel, etc.)
    void handle_special_request(TRANS_TYPE, PAYLD_TYPE);

    // Helper function to add requests to connectivity queue
    void b_add_request_to_conn_queue(TRANS_TYPE req);

    // Size of the queues
    unsigned int queue_size;



    // Two events reflecting the state of the new_requests queue
    sc_event new_requests_nonempty;
    sc_event new_requests_nonfull;
    sc_event conn_queue_nonfull;
    sc_event conn_queue_nonempty;
    sc_event finished_requests_nonempty;
    sc_event finished_requests_nonfull;
    

    // The queues
    bpq_t new_planning_requests;
    bpq_t connectivity_requests;
    bpq_t finished_planning_requests;
    bis_t temp_outgoing_requests;


#ifdef PROPERTY_2
    vect_mon_t* prop2;
#endif

  }; // flight_planner

}; //namespace

#endif
