#include "io_module.h"

using airline::io_module;


// The constructor
io_module::io_module(sc_module_name module_name, unsigned int q_size = 0 ) :
  sc_module(module_name),
  basic_module(),
  in_port_master("port_master_to_io"),
  test_out_port("test_out"),
  test_in_port("test_in"),
  clk("clock"),
  incoming_requests(bpq_t(q_size))
{
  
  SC_THREAD( send_new_requests );
  sensitive << clk;

  SC_THREAD ( get_requests_for_selection );
  sensitive << clk;
  
  test_in_port(*this);
  queue_size = q_size;


}

/**
 * The module is instantiated before the monitors are
 * instantiated. This is why we have to wait until the end of
 * elaboration to get a pointer to the monitors.
 */
void
io_module::before_end_of_elaboration() { 
#ifdef PROPERTY_2
  sc_module::before_end_of_elaboration();
  prop2 = sc_get_curr_simcontext()->get_observer()->get_vector_of_monitors_by_index(1);
#endif
}

/** 
 * Receive a transaction from the test generator. This is an
 * implementation of the function defined in the interface test_if
 * (which this io_module is implementing), and is necessary for the
 * sc_export through which we receive the test vectors.
 */
status_t io_module::receive_transaction(const token_t* t) {
  dprint(9, "io_module %s: receiving a transaction.\n", name());

  unsigned int request_id = t->get_request()->get_request_id();



  dprint(9, "io_module %s: transaction id is %d.\n", name(), request_id);
  
  // Check if we have enough storage in the queue
  unsigned int curr_size = incoming_requests.size();
  dprint(9, "io_module %s: incoming queue size is %d.\n", name(), curr_size);

  if (curr_size < queue_size) {
    incoming_requests.push(t->get_request());
    dprint(8, "io_module %s: successfully added request %d.\n", name(), request_id);
    inc_queue_nonempty.notify(SC_ZERO_TIME); //delta notification

#ifdef PROPERTY_2
    // Verification
    for(vect_mon_t::const_iterator it = prop2->begin(); it != prop2->end(); it++) {
      ((property2*) (*it))->io_module_receive_transaction(request_id);
    }
#endif

    return SUCCESS;
  }
  
  else {
    dprint(8, "io_module %s: incomding queue is full.\n", name());
    dprint(8, "io_module %s: Rejecting request %d.\n", name(), request_id);
    return ERROR;
  }
  
}




/**
 * Send new requests to the master_module for handling. Pauses in case
 * the channel is full, and resumes transfer when the channel can
 * accommodate another token.
 */
void io_module::send_new_requests() {
  while (true) {
    dprint(8, "io_module %s: checking the incoming queue for new requests\n", name());
    unsigned int queue_size = incoming_requests.size();
    dprint(9, "io_module %s: there are %d tokens in the incoming queue\n", name(), queue_size);
    
    if (queue_size > 0) {
      TRANS_TYPE curr_request = incoming_requests.toppop();
      dprint(9, "io_module %s: dequeued request %d\n", name(), curr_request->get_request_id());

      // Technically we don't need a while loop because we are the
      // only driver on this channel. However, here the loop forces us
      // to double check nb_can_put before we actually send the token.
      bool request_transmitted = false;
      while (! request_transmitted) {
	tlm_tag_t tag;
	if (out_port_master->nb_can_put( &tag )) {
	  dprint(9, "io_module %s: the channel to the master module has space for a token\n",  name());
	  dprint(9, "io_module %s: sending request %d to the master module\n", 
		 name(), curr_request->get_request_id());
	  
	  // Create a new token for this request
	  token_t* new_token = new token_t(curr_request, NEW);
	  out_port_master->nb_put(new_token);
	  request_transmitted = true;
#ifdef PROPERTY_2
	  for(vect_mon_t::const_iterator it = prop2->begin(); it != prop2->end(); it++) {
	    ((property2*) (*it))->io_module_send_to_master(curr_request->get_request_id());
	  }
#endif
	}

	else {
	  // The channel is currently full. Waiting until it is free again...
	  dprint(9, "io_module %s: channel to master is full.\n", name());
	  dprint(9, "io_module %s: Suspending send_new_request() until the channel becomes available.\n", name());
	  wait(out_port_master->ok_to_put(&tag));
	  dprint(9, "io_module %s: send_new_request has been awoken by a signal,\n", name());
	  dprint(9, "io_module %s: so the channel must be available right now.\n", name());
	} //else
      } // transmission loop
      
      dprint(9, "io_module %s: suspending send_new_request() until the next clock tick.\n", name());
      wait();
    } // incoming queue > 0
    
    else {
      dprint(9, "io_module %s: No new requests available.\n", name());
      dprint(9, "io_module %s: Suspending thread until there is a new incoming request.\n", name());
      wait(inc_queue_nonempty);
    } // incoming queue is empty
  }
}



/**
 * Receives completed requests from the master module
 */
void io_module::get_requests_for_selection() {
  dprint(4, "%s: %s() started\n", name(), __FUNCTION__);
  while(true) {
    dprint(4, "%s: %s() starts a new I/O receiving loop\n", name(), __FUNCTION__);
    dprint(5, "%s: checking for new requests on %s\n", name(), in_port_master.name());
      
    tlm_tag_t t;
    if (! in_port_master->nb_can_get( &t )) {
      dprint(5, "%s: no tokens available on %s\n", name(), in_port_master.name());
      dprint(6, "%s: suspending %s() until new tokens become available on %s\n", 
	     name(), __FUNCTION__, in_port_master.name());
      
      tlm_tag_t tag;
      wait(in_port_master->ok_to_get(&tag));
      dprint(6, "%s: %s() awoken; new tokens on %s\n", name(), __FUNCTION__, in_port_master.name());
    }
     
    dprint(5, "%s: token available on %s\n", name(), in_port_master.name());
    token_t* req = new token_t();
    in_port_master->nb_get(req);
    dprint(5, "%s: consumed request_id %d with status %s from %s\n", 
	   name(),
	   req->get_request()->get_request_id(), 
	   req->get_payload() == SELECT ? "SELECT" : "OTHER",
	   in_port_master.name());
    dprint(6, "%s: suspending %s() until the next clock cycle\n", name(), __FUNCTION__); 
    
#ifdef PROPERTY_2
    for(vect_mon_t::const_iterator it = prop2->begin(); it != prop2->end(); it++) {
      ((property2*) (*it))->io_module_receive_from_master(req->get_request()->get_request_id());
    }
#endif

    wait();
  } // while true
} // get_new_requests()
