#include "master_module.h"

using airline::master_module;


// The constructor
master_module::master_module(sc_module_name module_name, unsigned int q_size = 0 ) :
  sc_module(module_name),
  basic_module(),
  in_port_planner("port_from_planner"),
  in_port_io1("port_from_io1"),
  out_port_planner("port_to_planner"),
  out_port_io1("port_to_io"),
  clk("clock"),
  new_req(bpq_t(q_size)),
  plan_done(bpq_t(q_size)),
  queue_size(q_size)

{
  SC_THREAD( get_new_requests );
  sensitive << clk.pos();
  
  SC_THREAD( request_planning );
  sensitive << clk.pos();


  SC_THREAD( receive_planning );
  sensitive << clk.pos();

  SC_THREAD( send_planning );
  sensitive << clk.pos();
}


/**
 * The module is instantiated before the monitors are
 * instantiated. This is why we have to wait until the end of
 * elaboration to get a pointer to the monitors.
 */
void
master_module::before_end_of_elaboration() { 
#ifdef PROPERTY_2
  sc_module::before_end_of_elaboration();
  prop2 = sc_get_curr_simcontext()->get_observer()->get_vector_of_monitors_by_index(1);
#endif
}


/**
 * Receives new requests from the channel connected to the first IO module.
 */
void master_module::get_new_requests() {
  dprint(4, "%s: %s() started\n", name(), __FUNCTION__);
  while(true) {
    dprint(4, "%s: %s() starts a new I/O receiving loop\n", name(), __FUNCTION__);
    dprint(5, "%s: checking for new requests on %s\n", name(), in_port_io1.name());
      
    tlm_tag_t t;
    if (! in_port_io1->nb_can_get( &t )) {
      dprint(5, "%s: no tokens available on %s\n", name(), in_port_io1.name());
      dprint(6, "%s: suspending get_new_requests() until new tokens become available on %s\n", 
	     name(), in_port_io1.name());
      
      tlm_tag_t tag;
      wait(in_port_io1->ok_to_get(&tag));
      dprint(6, "%s: get_new_requests() awoken; new tokens on %s\n", name(), in_port_io1.name());
    }
     
    dprint(5, "%s: token available on %s\n", name(), in_port_io1.name());
    token_t* req = new token_t();
    in_port_io1->nb_get(req);
    dprint(5, "%s: consumed request_id %d with status %s from %s\n", 
	   name(),
	   req->get_request()->get_request_id(), 
	   req->get_payload() == NEW ? "NEW" : "OTHER",
	   in_port_io1.name());
    
#ifdef PROPERTY_2
    for(vect_mon_t::const_iterator it = prop2->begin(); it != prop2->end(); it++) {
      ((property2*) (*it))->master_receive_request_from_io(req->get_request()->get_request_id());
    }
#endif
    
    if (req->get_payload() == NEW) {
      unsigned int curr_size = new_req.size();
      dprint(5, "%s: queue new_req currently has %d items\n", name(), curr_size);
      
      if (curr_size >= queue_size) {
	dprint(6, "%s: new_req does not have space for a new token.\n", name());
	dprint(6, "%s: suspending get_new_requests() until new_req has space\n", 	name());
	wait(new_req_nonfull);
	dprint(6, "%s: get_new_requests() awoken; space available in new_req\n", name());
      }
	
      assert(new_req.size() < queue_size);
	
      dprint(5, "%s: adding request %d to internal queue \"new_req\"\n", 
	     name(), 
	     req->get_request()->get_request_id());
	
      new_req.push(req->get_request());
      new_req_nonempty.notify(SC_ZERO_TIME);

      dprint(5, "%s: mapping request %d to Q_NEW_REQ\n", name(), 
	     req->get_request()->get_request_id());
      stagemap[req->get_request()->get_request_id()] = Q_NEW_REQ;
      
      dprint(4, "%s: suspending %s() until the next clock tick\n", name(), __FUNCTION__);
      wait();
    } // it was a new request
    
    else {
      handle_special_request(req->get_request(), req->get_payload());
    }
  } // while true
} // get_new_requests()


/**
 * Receives requests from the planner and calculates the cost of the
 * trip.
 */
void master_module::receive_planning() {
  dprint(4, "%s: %s() started\n", name(), __FUNCTION__);

  while (true) {
    dprint(4, "%s: %s() started another receiving loop\n", name(), __FUNCTION__);
    dprint(5, "%s: checking for new requests on %s\n", name(), in_port_planner.name());
      
    tlm_tag_t t;
    if (! in_port_planner->nb_can_get( &t )) {
      dprint(6, "%s: no tokens available on %s\n", name(), in_port_planner.name());
      dprint(6, "%s: suspending receive_planning() until new tokens become available on %s\n", name(), in_port_planner.name());
      tlm_tag_t tag;
      wait(in_port_planner->ok_to_get(&tag));
      dprint(6, "%s: receive_planning() awoken; new tokens on %s\n", name(), in_port_planner.name());
    }
     
    token_t* req = new token_t();
    in_port_planner->nb_get(req);
    dprint(5, "%s: consumed request_id %d with status %s from %s\n", 
	   name(),
	   req->get_request()->get_request_id(), 
	   (req->get_payload() == PLANDONE) ? "PLANDONE" : (req->get_payload() == UPDATED ? "UPDATED" : "OTHER"),
	   in_port_planner.name());
    
    
#ifdef PROPERTY_2
    for(vect_mon_t::const_iterator it = prop2->begin(); it != prop2->end(); it++) {
      ((property2*) (*it))->master_receive_request_from_flight_planner(req->get_request()->get_request_id());
    }
#endif

    if (req->get_payload() == PLANDONE) {
      unsigned int curr_size = plan_done.size();
      dprint(6, "%s: queue plan_done currently has %d items\n", name(), curr_size);
      
      if (curr_size == queue_size) {
	dprint(6, "%s: plan_done does not have space for a new token.\n", name());
	dprint(6, "%s: suspending receive_planning() until plan_done has space\n", name());
	wait(plan_done_nonfull);
	dprint(6, "%s: receive_planning() awoken; space available in plan_done\n", name());
      }
      
      assert(plan_done.size() < queue_size);
      
      // Check for cancelations and preemptively drop the request
      // (rather than wait for the cancelation request to catch up
      if ( stagemap [req->get_request()->get_request_id()] == CANCELED ) {
	dprint(5, "%s: request %d has been canceled. Dropping the request.\n", name());
	stagemap.erase(stagemap.find(req->get_request()->get_request_id()));
	dprint(5, "%s: Deleted the entry in the stagemap\n", name());
	dprint(5, "%s: Suspending %s until the next clock tick\n", name(), __FUNCTION__);
	wait();
      } // canceled request is dropped

      else { // active request is propagated
	assert (stagemap[req->get_request()->get_request_id()] == PLANNING );
	dprint(6, "%s: adding request %d (%d) to internal queue \"plan_done\"\n",
	       name(),
	       req->get_request()->get_request_id(),
	       req->get_request()->get_req_sequence());
	
	plan_done.push(req->get_request());
	
	dprint(5, "%s: mapping request %d to Q_PLAN_DONE\n", name(), 
	       req->get_request()->get_request_id());
	stagemap[req->get_request()->get_request_id()] = Q_PLAN_DONE;
	plan_done_nonempty.notify(SC_ZERO_TIME);
	  
      } // not a canceled request
    } // it was a standard request
    
    else {
      handle_special_request(req->get_request(), req->get_payload());
    }

    dprint(4, "%s: suspending %s() until the next clock tick\n", name(), __FUNCTION__);
    wait();
  } // receiving loop
} // receive_planning()


/**
 * Send back to IO module the completed request for approval
 */
void master_module::send_planning() {
  dprint(4, "%s: %s() started\n", name(), __FUNCTION__);
  

  while (true) {
    dprint(5, "%s: checking the plan_done queue for new requests\n", name());
    unsigned int queue_size = plan_done.size();
    dprint(9, "%s: there are %d tokens in the queue\n", name(), queue_size);
    
    if (queue_size > 0) {
      TRANS_TYPE curr_request = plan_done.toppop();
      dprint(6, "%s: dequeued request %d\n", name(), curr_request->get_request_id());

      // First check if we need to keep this request around
      if (curr_request->get_speculative_status()) {
	dprint(6, "%s: this request is speculative. It goes back to the queue.\n", name());
	plan_done.push(curr_request);
	dprint(6, "%s: Suspending thread until somebody notifies plan_done_nonempty.\n", name());
	wait(plan_done_nonempty);
      }
      
      else {
	
	// Technically we don't need a while loop because we are the
	// only driver on this channel. However, here the loop forces us
	// to double check nb_can_put before we actually send the token.
	bool request_transmitted = false;
	while (! request_transmitted) {
	  tlm_tag_t tag;
	  if (out_port_io1->nb_can_put( &tag )) {
	    dprint(6, "%s: the channel to the io1 module has space for a token\n",  name());
	    dprint(5, "%s: sending request %d to the io module\n", 
		   name(), curr_request->get_request_id());
	  
	    // Create a new token for this request
	    token_t* new_token = new token_t(curr_request, SELECT);
	    out_port_io1->nb_put(new_token);
	    request_transmitted = true;

#ifdef PROPERTY_2
	    for(vect_mon_t::const_iterator it = prop2->begin(); it != prop2->end(); it++) {
	      ((property2*) (*it))->master_send_request_to_io_module(curr_request->get_request_id());
	    }
#endif

	    // Also update the map so that we know where the request went...
	    stagemap[curr_request->get_request_id()] = RETURNED;
	  }

	  else {
	    // The channel is currently full. Waiting until it is free again...
	    dprint(6, "%s: channel to master is full.\n", name());
	    dprint(6, "%s: Suspending send_new_request() until the channel becomes available.\n", name());
	    wait(out_port_io1->ok_to_put(&tag));
	    dprint(6, "%s: send_new_request has been awoken by a signal,\n", name());
	    dprint(6, "%s: so the channel must be available right now.\n", name());
	  } //else
	} // transmission loop
	
	dprint(9, "%s: suspending %s() until the next clock tick.\n", name(), __FUNCTION__);
	wait();
      } // This is not a speculative request
    } // incoming queue > 0

    else {
      dprint(6, "%s: Nothing available in plan_done\n", name());
      dprint(6, "%s: Suspending thread until there is a new request.\n", name());
      wait(plan_done_nonempty);
    } // plan_done queue is empty
  }

  dprint(4, "%s: %s() finished\n", name(), __FUNCTION__);
}



/**
 * Sends a request to the flight planner requesting a plan for the
 * flights
 */
void master_module::request_planning() {
  dprint(4, "%s: %s() started\n", name(), __FUNCTION__);
  while (true) {
    dprint(4, "%s: %s() start a new requesting loop\n", name(), __FUNCTION__);

    unsigned int queue_size = new_req.size();
    dprint(5, "%s: there are %d tokens in new_req \n", name(), queue_size);
    
    if (queue_size == 0) {
      dprint(6, "%s: Suspending request_planning() until there new_req is nonempty.\n", name());
      wait(new_req_nonempty);
      dprint(6, "%s: request_planning() was awoken; new_req must have something\n", name());
    }
    
    assert (new_req.size() > 0 );
    
    TRANS_TYPE curr_request = new_req.toppop();
    new_req_nonfull.notify(SC_ZERO_TIME);
      
    dprint(5, "%s: dequeued request %d\n", name(), curr_request->get_request_id());

    tlm_tag_t tag;
    if (! out_port_planner->nb_can_put( &tag )) {
      dprint(5, "%s: Suspending request_planning() until the channel becomes available.\n", name());
      wait(out_port_planner->ok_to_put(&tag));
      dprint(5, "%s: request_planning() has been awoken; channel to planner must be available.\n", name());
    }

    // Create a new token for this request
    token_t* new_token = new token_t(curr_request, PLAN);
    out_port_planner->nb_put(new_token);

#ifdef PROPERTY_2
    for(vect_mon_t::const_iterator it = prop2->begin(); it != prop2->end(); it++) {
      ((property2*) (*it))->master_send_to_flight_planner(curr_request->get_request_id());
    }
#endif
    
    dprint(5, "%s: mapping request %d to PLANNING\n", name(), 
	   curr_request->get_request_id());
    stagemap[curr_request->get_request_id()] = PLANNING;
    
    dprint(4, "%s: suspending request_planning() until the next clock tick.\n", name());
    wait();
  }
}

bool master_module::obj_eq(const TRANS_TYPE o1, const TRANS_TYPE o2) {
  return o1->get_request_id() == o2->get_request_id();
}

void master_module::print_function(const TRANS_TYPE obj) {
  printf("Considering transaction (id %d, leg %d)\n", obj->get_request_id(), obj->get_req_sequence());
}


/**
 * Handles special requests (eg. canceling or modifying a pending request.
 */
void master_module::handle_special_request(TRANS_TYPE req, PAYLD_TYPE stat) {
  dprint(9, "%s: %s().\n", name(), __FUNCTION__);
  
  int req_id = req->get_request_id();
  dprint(9, "%s: Request %d, action %s\n", name(), req_id, 
	 basic_protocol::basic_request<TRANS_TYPE,PAYLD_TYPE>::to_string(stat));
 
  // Find the stage of the request
  switch (stagemap[req_id] ) {
  case Q_NEW_REQ : 
    dprint(9, "%s: request %d is in new_req queue\n", name(), req_id);

    switch (stat) {
    case CANCEL:{
      int remove_status = new_req.remove(req, obj_eq);
      dprint(9, "%s: status of the remove is %d (0=success, -1 = not found)\n", name(), remove_status);
      break;
    }

    case UPDATED: {
      int update_status = new_req.update(req, obj_eq);
      dprint(9, "%s: status of the update is %d (0=success, -1 = not found)\n", name(), update_status);
      break;
    }
      
    default:
      cerr << "Unexpected action for this request." << endl;
      exit(1);
      break;
    }
    break;
    
  case PLANNING:
    dprint(9, "%s: request %d is in the planner\n", name());

    switch (stat) {
    case CANCEL: {
      int remove_status = new_req.remove(req, obj_eq);
      dprint(9, "%s: status of the remove is %d (0=success, -1 = not found)\n", name(), remove_status);
      break;
    }
    case UPDATED: {
      int update_status = new_req.update(req, obj_eq);
      dprint(9, "%s: status of the update is %d (0=success, -1 = not found)\n", name(), update_status);
      break;
    }
      
    default:
      cerr << "Unexpected action for this request." << endl;
      exit(1);
      break;
    }

    break;

  case Q_PLAN_DONE:
    dprint(6, "%s: request %d is in the plan_done queue\n", name(), req->get_request_id());
    
    dprint(6, "%s: dequeueing the request\n", name());
    TRANS_TYPE old_req = plan_done.extract(req, obj_eq, print_function);

    assert (old_req != NULL);
    
    dprint(6, "A printout of the old request follows\n");
    CONDPRINT(6, old_req->print_me());

    old_req->set_connecting_flights2(req->get_connecting_flights());
    old_req->set_speculative_bit_to(false);;
    
    dprint(6, "%s: queueing the updated request\n", name());
    plan_done.push(old_req);
    plan_done_nonempty.notify(SC_ZERO_TIME);

    dprint(6, "A printout of the new request follows\n");
    CONDPRINT(6, old_req->print_me());

    break;

  case CANCELED:
    dprint(9, "%s: request %d has been canceled previously\n", name());
    exit(1);
    break;

  case RETURNED:
    dprint(6, "%s: request %d has been returned to the IO module\n", name());
    exit(2);
    break;

  default:
    fprintf(stderr, "%s: handle_special_request(): request %d, status %s. Unexpected stage\n",  name(), req_id, stat);
  }
  
}

