#ifndef BOUNDED_QUEUE_H
#define BOUNDED_QUEUE_H

#include <queue>

namespace user {

  template < typename OBJ_TYPE, typename comp >
  class bounded_queue : 
    public std::priority_queue< OBJ_TYPE ,                 // Type of objects
			 std::vector< OBJ_TYPE >,   // Underlying container
			 comp >    {                // comparison function
    
  public:
    
    // Constructor
    bounded_queue(unsigned int s = 0):
      std::priority_queue<OBJ_TYPE, std::vector<OBJ_TYPE>, comp>(),
      q_size(s) 
    {}
    
    // Standard setter
    void set_size(unsigned int new_size) {
      q_size = new_size;
    }





    /**
     * Removes the element at the top of the queue and returns it to
     * the caller. Precondition: the queue is not empty.
     */
    OBJ_TYPE toppop() {
      assert( this->size() > 0 );

      OBJ_TYPE o = this->top();
      this->pop();
      return o;
    }






    /**
     * Adds elements to the queue, making sure that the number of
     * elements in the queue is less than the starting size as
     * specified by the user.
     */
    void push(OBJ_TYPE o) {
        
      unsigned int curr_size = this->size();
      if (curr_size >= q_size) {
	fprintf(stderr, "ERROR: Bounded queue: queue size is %d and there are %d elements already in the queue.\n", q_size, curr_size);
	fprintf(stderr, "ERROR: Attempting to add another object fails in bounded_queue.cc\n");
	exit(1);
      }
      else {
	std::priority_queue<OBJ_TYPE , std::vector< OBJ_TYPE >, comp>::push(o);
      }
    }
    



    
    /**
     * Removes an object from the queue. Equality of objects is
     * determined by the supplied function @obj_equal. Returns 0 on
     * success, -1 if the object could not be found.
     */
    int remove(OBJ_TYPE reference, bool (*obj_equal) (OBJ_TYPE, OBJ_TYPE) ) {
      
      if (this->empty()) {
	cerr << "Calling remove on an empty queue. Something is fishy here..." << endl;
	return -1;
      }
      
      int value_to_return = -1;
      
      std::priority_queue< OBJ_TYPE , std::vector< OBJ_TYPE >, comp > copy_q;
      while (! this->empty()) {
	OBJ_TYPE current = this->top();
	
	if (! obj_equal(current, reference)) {
	  copy_q.push(current);
	}

	else {
	  value_to_return = 0;
	}

	this->pop();
      }

      // Return all objects back to the original queue
      while (! copy_q.empty()) {
	OBJ_TYPE current = copy_q.top();
	this->push(current);
	copy_q.pop();
      }

      return value_to_return;
    }



    /**
     * Removes an object from the queue and returns it to the
     * caller. Equality of objects is determined by the supplied
     * function @obj_equal. Returns the object on success, null if the
     * object could not be found.
     */
    OBJ_TYPE extract(OBJ_TYPE reference, bool (*obj_equal) (OBJ_TYPE, OBJ_TYPE), void (*print_me) (OBJ_TYPE) ) {
      
      if (this->empty()) {
	cerr << "Calling extract on an empty queue. Something is fishy here..." << endl;
	return NULL;
      }
      
      OBJ_TYPE value_to_return = NULL;
      
      std::priority_queue< OBJ_TYPE , std::vector< OBJ_TYPE >, comp > copy_q;
      while (! this->empty()) {
	OBJ_TYPE current = this->top();
	
	if (! obj_equal(current, reference)) {
	  copy_q.push(current);
	  print_me(current);
	}

	else {
	  value_to_return = current;
	}

	this->pop();
      }

      // Return all objects back to the original queue
      while (! copy_q.empty()) {
	OBJ_TYPE current = copy_q.top();
	this->push(current);
	copy_q.pop();
      }

      return value_to_return;
    }


    
    /**
     * Updates an object in the queue. Equality of objects is
     * determined by the supplied function @obj_equal. Returns 0 on
     * success, -1 if the object could not be found.
     */
    int update(OBJ_TYPE reference, bool (*obj_equal) (OBJ_TYPE, OBJ_TYPE) ) {
      
      if (this->empty()) {
	cerr << "Calling update on an empty queue. Something is fishy here..." << endl;
	return -1;
      }
      
      int value_to_return = -1;
      
      std::priority_queue< OBJ_TYPE , std::vector< OBJ_TYPE >, comp > copy_q;
      while (! this->empty()) {
	OBJ_TYPE current = this->top();
	
	if (! obj_equal(current, reference)) {
	  copy_q.push(current);
	}

	else {
	  value_to_return = 0;
	  copy_q.push(reference);
	}

	this->pop();
      }

      // Return all objects back to the original queue
      while (! copy_q.empty()) {
	OBJ_TYPE current = copy_q.top();
	this->push(current);
	copy_q.pop();
      }

      return value_to_return;
    }





    
  protected:
    unsigned int q_size; // Size of the queue
    
  }; // bounded_queue
}; // namespace

#endif
